import glob
import json
import os
import pathlib
import re
import reprlib
import shutil
import weakref
import zipfile

from collections import defaultdict
from fnmatch import fnmatch, translate
from io import TextIOWrapper
from pathlib import Path
from pprint import pprint
from warnings import warn

from scitrack import get_text_hexdigest
from tinydb import Query, TinyDB
from tinydb.middlewares import CachingMiddleware
from tinydb.storages import JSONStorage

from cogent3.app.data_store_new import (
    get_data_source,
    load_record_from_json,
    make_record_for_json,
)
from cogent3.util.deserialise import deserialise_not_completed
from cogent3.util.io import atomic_write, get_format_suffixes, open_
from cogent3.util.misc import extend_docstring_from
from cogent3.util.parallel import is_master_process
from cogent3.util.table import Table
from cogent3.util.union_dict import UnionDict


__author__ = "Gavin Huttley"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Gavin Huttley"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Alpha"


# handling archive, member existence
SKIP = "skip"
OVERWRITE = "overwrite"
RAISE = "raise"
IGNORE = "ignore"


class DataStoreMember(str):  # pragma: no cover
    def __new__(klass, name, parent=None, id=None):
        from cogent3.util.warning import deprecated

        deprecated(
            "class",
            "DataStoreMember",
            "cogent3.data_store_new.DataMember",
            "2023.3",
            "use cogent3.data_store_new.DataMember",
        )

        result = str.__new__(klass, name)
        result.name = os.path.basename(name)
        result.parent = parent
        result._file = None
        result.id = id
        return result

    def read(self):
        """returns contents"""
        return self.parent.read(self)

    def open(self):
        """returns file-like object"""
        if self._file is None:
            self._file = self.parent.open(self.name)
        return self._file

    def close(self):
        """closes file"""
        if self._file is None:
            return
        self._file.close()
        self._file = None

    @property
    def md5(self):
        return self.parent.md5(self, force=True)


class ReadOnlyDataStoreBase:  # pragma: no cover

    store_suffix = None

    def __init__(self, source, suffix=None, limit=None, verbose=False, md5=True):
        """
        Parameters
        ----------
        source
            path to directory / zip file. Forced to end with store_suffix.
        suffix
            only members whose name matches the suffix are considered included
        limit
            the maximum number of members to consider
        verbose
            displays files that don't match search (applies only to the Zipped
            variant)
        md5 : bool
            record md5 hexadecimal checksum of read data when possible
        """
        # assuming delimiter is /

        from cogent3.util.warning import deprecated

        deprecated(
            "class",
            "ReadOnlyDataStoreBase",
            "cogent3.data_store_new.DataStoreDirectory",
            "2023.3",
            "use cogent3.data_store_new.DataStoreABC",
        )

        # todo this approach to caching persistent arguments for reconstruction
        # is fragile. Need an inspect module based approach
        d = locals()
        self._persistent = UnionDict({k: v for k, v in d.items() if k != "self"})

        source = str(source)
        suffix = suffix or ""
        if suffix != "*":  # wild card search for all
            suffix = re.sub(r"^[\s.*]+", "", suffix)  # tidy the suffix
        source = re.sub(r"/+$", "", source)  # tidy the source

        self.suffix = suffix
        if self.store_suffix and not source.endswith(self.store_suffix):
            source = ".".join([source, self.store_suffix])
        self.source = str(pathlib.Path(source).expanduser())
        self.mode = "r"
        self._members = []
        self.limit = limit
        self._verbose = verbose
        self._md5 = md5
        self._checksums = {}

    def __getstate__(self):
        return self._persistent.copy()

    def __setstate__(self, data):
        new = self.__class__(**data)
        self.__dict__.update(new.__dict__)
        return self

    def __repr__(self):
        if len(self) > 3:
            sample = str(list(self[:3]))
            sample = f"{sample[:-1]}..."
        else:
            sample = list(self)

        num = len(self)
        name = self.__class__.__name__
        return f"{num}x member {name}(source='{self.source}', members={sample})"

    def __str__(self):
        return str(list(self))

    def head(self, n=5):
        """displays top n members"""
        pprint(self[:n])

    def tail(self, n=5):
        """displays last n members"""
        pprint(self[-n:])

    def __iter__(self):
        for i, member in enumerate(self.members):
            if not isinstance(member, DataStoreMember):
                member = DataStoreMember(self.get_absolute_identifier(member), self)
                self.members[i] = member
            yield member

    def __getitem__(self, index):
        return self.members[index]

    def __len__(self):
        return len(self.members)

    def __contains__(self, identifier):
        """whether relative identifier has been stored"""
        if isinstance(identifier, DataStoreMember):
            return identifier.parent is self

        if not identifier.endswith(self.suffix):
            suffix = pathlib.Path(identifier).suffix
            # possible an "added" file
            if self.store_suffix == "zip":
                klass = ReadOnlyZippedDataStore
            else:
                klass = ReadOnlyDirectoryDataStore
            new = klass(self.source, suffix=suffix)
            return identifier in new
        identifier = self.get_relative_identifier(identifier)
        result = False
        for member in self.members:
            if identifier in member:
                result = True
                break
        return result

    def get_member(self, identifier):
        """returns DataStoreMember"""
        identifier = self.get_relative_identifier(identifier)
        for member in self.members:
            if identifier in member:
                return member
        return None

    def get_relative_identifier(self, identifier):
        """returns the identifier relative to store root path"""
        if isinstance(identifier, DataStoreMember) and identifier.parent is self:
            return identifier

        source = self.source
        identifier = os.path.basename(identifier)
        if source.endswith(".zip"):
            # we insert the source path into identifier name
            # for zip members to ensure inflation creates a directory
            # containing them
            source = source.replace(".zip", "")
            source = os.path.basename(source)
            identifier = f"{source}{os.sep}{identifier}"
        else:
            identifier = Path(identifier)
            identifier = identifier.name

        return identifier

    def get_absolute_identifier(self, identifier, from_relative=False):
        """returns the identifier relative to the root path"""
        if not from_relative:
            identifier = self.get_relative_identifier(identifier)
        source = self.source.replace(".zip", "")
        if isinstance(identifier, DataStoreMember):
            identifier = identifier.name
        elif not identifier.startswith(source):
            identifier = f"{source}{os.sep}{identifier}"
        return identifier

    def read(self, identifier):
        """reads data corresponding to identifier"""
        if isinstance(identifier, DataStoreMember) and identifier.parent is self:
            identifier = identifier.name
        source = self.open(identifier)

        data = source.read()
        if self._md5:
            self._checksums[identifier] = get_text_hexdigest(data)
        source.close()
        return data

    @property
    def members(self):
        raise NotImplementedError  # override in subclasses

    def open(self, identifier):
        raise NotImplementedError

    def filtered(self, pattern=None, callback=None):
        """returns list of members for which callback returns True"""
        assert any([callback, pattern]), "Must provide a pattern or a callback"
        if pattern:
            result = [m for m in self if fnmatch(m, pattern)]
        else:
            result = [m for m in self if callback(m)]
        return result

    def md5(self, identifier, force=True):
        """
        Parameters
        ----------
        identifier
            name of data store member
        force : bool
            forces reading of data if not already done

        Returns
        -------
        md5 checksum for the member, if available, None otherwise
        """
        md5_setting = self._md5  # for restoring automatic md5 calc setting
        absoluteid = self.get_absolute_identifier(identifier)
        if force and absoluteid not in self._checksums:
            self._md5 = True
            _ = self.read(absoluteid)

        result = self._checksums.get(absoluteid, None)
        self._md5 = md5_setting
        return result

    def write_log(self, unique_id: str, data: str) -> None:
        _ = self.write(unique_id, data)


class ReadOnlyDirectoryDataStore(ReadOnlyDataStoreBase):  # pragma: no cover
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from cogent3.util.warning import deprecated

        deprecated(
            "class",
            f"{self.__class__.__name__}",
            "DataStoreDirectory",
            "2023.3",
            "use cogent3.open_data_store",
        )

    @property
    def members(self):
        if not self._members:
            pattern = f"{self.source}/**/*.{self.suffix}"
            paths = glob.iglob(pattern, recursive=True)
            members = []
            for i, path in enumerate(paths):
                if self.limit and i >= self.limit:
                    break
                member = DataStoreMember(self.get_absolute_identifier(path), self)
                members.append(member)
            self._members = members
        return self._members

    def open(self, identifier):
        identifier = self.get_absolute_identifier(identifier, from_relative=False)
        if not os.path.exists(identifier):
            raise ValueError(f"path '{identifier}' does not exist")

        return open_(identifier)


class SingleReadDataStore(ReadOnlyDirectoryDataStore):  # pragma: no cover
    """simplified for a single file"""

    def __init__(self, source, *args, **kwargs):
        """
        Parameters
        source
            path to one file
        args
            ignored
        kwargs
            ignored
        """
        path = Path(source).expanduser()
        assert path.exists() and path.is_file()
        super(SingleReadDataStore, self).__init__(
            str(path.parent), suffix=str(path.suffix)
        )
        self._members = [DataStoreMember(path, self)]


class ReadOnlyZippedDataStore(ReadOnlyDataStoreBase):  # pragma: no cover
    store_suffix = "zip"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        from cogent3.util.warning import discontinued

        discontinued(
            "class",
            "ReadOnlyZippedDataStore",
            "2022.3",
            "dropping support for zip archives",
        )

    @property
    def members(self):
        if os.path.exists(self.source) and not self._members:
            source_path = self.source.replace(Path(self.source).suffix, "")
            pattern = f"*.{self.suffix}"
            members = []
            with zipfile.ZipFile(self.source) as archive:
                names = archive.namelist()
                num_matches = 0
                for name in names:
                    name = os.path.basename(name)
                    if fnmatch(name, pattern):
                        num_matches += 1
                        member = DataStoreMember(os.path.join(source_path, name), self)
                        members.append(member)
                    elif self._verbose:
                        print(f"Did not match {name}")

                    if self.limit and num_matches >= self.limit:
                        break
            self._members = members

        return self._members

    def open(self, identifier):
        identifier = self.get_relative_identifier(identifier)
        archive = zipfile.ZipFile(self.source)
        record = archive.open(identifier.replace("\\", "/"))
        record = TextIOWrapper(record, encoding="latin-1")
        return record


class WritableDataStoreBase:  # pragma: no cover
    """a writeable data store"""

    def __init__(self, if_exists=RAISE, create=False):
        """
        if_exists : str
             behaviour when the destination already exists. Valid constants are
             defined in this file as OVERWRITE, RAISE, IGNORE (they
             correspond to lower case version of the same word)
        create : bool
            if True, the destination is created
        """
        from cogent3.util.warning import deprecated

        deprecated(
            "class",
            "WritableDataStoreBase",
            "cogent3.data_store_new.DataStoreDirectory",
            "2023.3",
            "use cogent3.data_store_new.DataStoreABC",
        )

        d = locals()
        d = UnionDict({k: v for k, v in d.items() if k != "self"})
        if self._persistent:
            self._persistent |= d
        else:
            self._persistent = d

        self._members = []
        if_exists = if_exists.lower()
        assert if_exists in (OVERWRITE, SKIP, RAISE, IGNORE)
        self._source_create_delete(if_exists, create)

    def make_relative_identifier(self, data):
        """returns identifier for a new member relative to source"""

        if isinstance(data, DataStoreMember):
            data = data.name
        elif type(data) != str:
            data = get_data_source(data)
            if data is None:
                raise ValueError(
                    "objects for storage require either a "
                    "source or info.source string attribute"
                )
        basename = os.path.basename(data)
        suffix, comp = get_format_suffixes(basename)
        if suffix and comp:
            pattern = f".{suffix}.{comp}$"
        elif suffix:
            pattern = f".{suffix}$"
        elif comp:
            pattern = f".{comp}*$"
        else:
            pattern = None
        if pattern:
            basename = re.sub(pattern, "", basename)
        basename = f"{basename}.{self.suffix}"
        return basename

    def make_absolute_identifier(self, data):
        """returns a absolute identifier for a new member, includes source"""
        basename = self.make_relative_identifier(data)
        return self.get_absolute_identifier(basename, from_relative=True)

    def add_file(self, path, make_unique=True, keep_suffix=True, cleanup=False):
        """
        Parameters
        ----------
        path : str
            location of file to be added to the data store
        keep_suffix : bool
            new path will retain the suffix of the provided file
        make_unique : bool
            a successive number will be added to the name before the suffix
            until the name is unique
        cleanup : bool
            delete the original
        """
        relativeid = self.make_relative_identifier(path)
        relativeid = Path(relativeid)
        path = Path(path)
        if keep_suffix:
            relativeid = str(relativeid).replace(
                relativeid.suffix, "".join(path.suffixes)
            )
            relativeid = Path(relativeid)

        suffixes = "".join(relativeid.suffixes)
        new = relativeid
        num = 0
        while True:
            if not str(relativeid) in self:
                if num:
                    new = str(relativeid).replace(suffixes, f"-{num}{suffixes}")
                break

            num += 1
        relativeid = new
        data = SingleReadDataStore(path)[0].read()
        self.write(str(relativeid), data)

        if cleanup:
            path.unlink()

        return relativeid

    def write_incomplete(self, identifier, not_completed) -> DataStoreMember:
        """

        Parameters
        ----------
        identifier : str
            identifier for record
        not_completed : NotComplete
            instance that records key details for why incomplete
        Returns
        -------
        None if storage class does not support writing incomplete, otherwise
        a DataStoreMember.
        """
        if self.suffix != "json":
            msg = f"not supported for {self.__class__.__name__}"
            warn(msg, UserWarning)
            return

        record = make_record_for_json(identifier, not_completed, False)
        record = json.dumps(record)
        return self.write(identifier, record)

    def write(self, identifier, data, *args, **kwargs) -> DataStoreMember:
        """
        Parameters
        ----------
        identifier : str
            identifier that data will be saved under. Must have a suffix matching
            self.suffix or ``.log``.
        data
            data to be saved. If a tinydb, must be an object that can be
            converted to json, or has a to_json() method. Otherwise, it must be a string.

        Returns
        -------
        DataStoreMember instance
        """
        if not isinstance(data, str):
            raise TypeError(f"data must be a string type, not {type(data)}")

        id_suffix = identifier.split(".")[-1]
        if id_suffix not in (self.suffix, "log"):
            raise ValueError(
                f"identifier does not end with required suffix {self.suffix}"
            )

    def close(self):
        pass


class WritableDirectoryDataStore(
    ReadOnlyDirectoryDataStore, WritableDataStoreBase
):  # pragma: no cover
    @extend_docstring_from(ReadOnlyDirectoryDataStore.__init__, pre=False)
    @extend_docstring_from(WritableDataStoreBase.__init__, pre=False)
    def __init__(
        self,
        source,
        suffix,
        mode="w",
        if_exists=RAISE,
        create=False,
        md5=True,
        **kwargs,
    ):
        """
        md5 : bool
            record md5 hexadecimal checksum of data when possible
        mode : str
            file opening mode, defaults to write
        """
        assert "w" in mode or "a" in mode
        ReadOnlyDirectoryDataStore.__init__(self, source=source, suffix=suffix, md5=md5)
        WritableDataStoreBase.__init__(self, if_exists=if_exists, create=create)
        from cogent3.util.warning import deprecated

        deprecated(
            "class",
            f"{self.__class__.__name__}",
            "DataStoreDirectory",
            "2023.3",
            "convert to sqlitedb using cogent3.app.data_store_new.convert_tinydb_to_sqlite",
        )

        d = locals()
        self._persistent = {k: v for k, v in d.items() if k != "self"}
        self.mode = mode

    def _has_other_suffixes(self, path, suffix):
        p = Path(path)
        allowed = {str(suffix).lower(), "log"}
        for f in p.iterdir():
            if get_format_suffixes(str(f))[0] not in allowed:
                return True
        return False

    def _source_create_delete(self, if_exists, create):
        if not is_master_process():
            return

        path = Path(self.source)
        if path.exists() and if_exists == RAISE:
            raise FileExistsError(f"'{self.source}' exists")
        elif path.exists() and if_exists == OVERWRITE:
            if self._has_other_suffixes(self.source, self.suffix):
                raise RuntimeError(
                    f"Unsafe to delete {self.source!r} as it contains "
                    f"files other than .{self.suffix} or .log files."
                    " You will need to remove this directly yourself."
                )
            try:
                shutil.rmtree(self.source)
            except NotADirectoryError:
                os.remove(self.source)
        elif not path.exists() and not create:
            raise FileNotFoundError(f"'{self.source}' does not exist")

        if create:
            path.mkdir(parents=True, exist_ok=True)

    @extend_docstring_from(WritableDataStoreBase.write)
    def write(self, identifier, data):
        if not data:
            return data

        super().write(identifier, data)
        id_suffix = identifier.split(".")[-1]
        if id_suffix not in (self.suffix, "log"):
            raise ValueError(
                f"identifier does not end with required suffix {self.suffix}"
            )

        relative_id = self.get_relative_identifier(identifier)
        absolute_id = self.get_absolute_identifier(relative_id, from_relative=True)

        if self._md5:
            self._checksums[absolute_id] = get_text_hexdigest(data)

        with atomic_write(str(absolute_id), in_zip=False) as out:
            out.write(data)

        member = DataStoreMember(relative_id, self)
        if relative_id not in self and relative_id.endswith(self.suffix):
            self._members.append(member)

        return member


def _db_lockid(path):
    """returns value for pid in LOCK record or None"""
    if not os.path.exists(path):
        return None

    with TinyDB(path) as db:
        query = Query().identifier.matches("LOCK")
        got = db.get(query)
        lockid = None if not got else got["pid"]

    return lockid


class ReadOnlyTinyDbDataStore(ReadOnlyDataStoreBase):  # pragma: no cover
    """A TinyDB based json data store"""

    store_suffix = "tinydb"

    @extend_docstring_from(ReadOnlyDirectoryDataStore.__init__)
    def __init__(self, *args, **kwargs):
        from cogent3.util.warning import discontinued

        discontinued(
            "class", "ReadOnlyTinyDbDataStore", "2022.3", "use cogent3.open_data_store"
        )

        kwargs["suffix"] = "json"
        super(ReadOnlyTinyDbDataStore, self).__init__(*args, **kwargs)
        self._db = None
        self._finish = None

    def __contains__(self, identifier):
        """whether identifier has been stored here"""
        if isinstance(identifier, DataStoreMember):
            return identifier.parent is self

        query = Query().identifier.matches(identifier)
        return self.db.contains(query)

    def __repr__(self):
        txt = super().__repr__()
        query = Query().completed == False
        num = self.db.count(query)
        if num > 0:
            txt = f"{txt}, {num}x incomplete"
        return txt

    @property
    def db(self):
        if self._db is None:
            storage = CachingMiddleware(JSONStorage)
            storage.WRITE_CACHE_SIZE = 50  # todo support for user specifying
            self._db = TinyDB(self.source, storage=storage)
            name = self.__class__.__name__
            if "readonly" in name.lower():
                # remove interface for inserting records making this a read only db
                self._db.insert = None
            else:
                self.lock()

            self._finish = weakref.finalize(self, self._close, self._db)

        return self._db

    def __del__(self):
        self.close()

    @classmethod
    def _close(cls, db):
        try:
            db.storage.flush()
            db.close()
        except ValueError:
            # file probably already closed
            pass

    def close(self):
        """closes the data store"""
        try:
            self.unlock()
            self.db.storage.flush()
        except ValueError:
            # file probably already closed
            pass
        self._finish()
        self._finish.detach()

    def lock(self):
        """if writable, and not locked, locks the database to this pid"""
        if not self.locked:
            self._db.insert(dict(identifier="LOCK", pid=os.getpid()))
            self._db.storage.flush()

    @property
    def locked(self):
        """returns lock pid or None if unlocked or pid matches self"""
        return _db_lockid(self.source) is not None

    def unlock(self, force=False):
        """remove a lock if pid matches. If force, ignores pid."""
        if "readonly" in self.__class__.__name__:
            # not allowed to touch a lock
            return

        query = Query().identifier.matches("LOCK")
        got = self.db.get(query)
        if not got:
            return

        lock_id = got["pid"]
        if lock_id == os.getpid() or force:
            self.db.remove(query)
            self.db.storage.flush()

        return lock_id

    @property
    def incomplete(self):
        """returns database records with completed=False"""
        query = Query().completed == False
        incomplete = []
        for record in self.db.search(query):
            member = DataStoreMember(record["identifier"], self, id=record.doc_id)
            incomplete.append(member)
        return incomplete

    @property
    def summary_incomplete(self):
        """returns a table summarising incomplete results"""
        # detect last exception line
        err_pat = re.compile(r"[A-Z][a-z]+[A-Z][a-z]+\:.+")
        types = defaultdict(list)
        indices = "type", "origin"
        for member in self.incomplete:
            record = member.read()
            record = deserialise_not_completed(record)
            key = tuple(getattr(record, k, None) for k in indices)
            match = err_pat.findall(record.message)
            types[key].append([match[-1] if match else record.message, record.source])

        header = list(indices) + ["message", "num", "source"]
        rows = []
        maxtring = reprlib.aRepr.maxstring
        reprlib.aRepr.maxstring = 45

        for record in types:
            messages, sources = list(zip(*types[record]))
            messages = reprlib.repr(
                ", ".join(m.splitlines()[-1] for m in set(messages))
            )
            sources = reprlib.repr(", ".join(s.splitlines()[-1] for s in sources))
            row = list(record) + [
                messages,
                len(types[record]),
                sources,
            ]
            rows.append(row)

        reprlib.aRepr.maxstring = maxtring  # restoring original val

        return Table(header=header, data=rows, title="incomplete records")

    @property
    def members(self):
        if not self._members:
            pattern = translate(f"*.{self.suffix}") if self.suffix else translate("*")
            members = []
            query = Query()
            query = (query.identifier.matches(pattern)) & (query.completed == True)
            for record in self.db.search(query):
                member = DataStoreMember(record["identifier"], self, id=record.doc_id)
                members.append(member)

                if self.limit and len(members) >= self.limit:
                    break
            self._members = members

        return self._members

    @extend_docstring_from(ReadOnlyDataStoreBase.get_absolute_identifier, pre=True)
    def get_absolute_identifier(self, identifier, from_relative=True):
        """For tinydb, this is the same as the relative identifier"""
        return self.get_relative_identifier(identifier)

    @extend_docstring_from(ReadOnlyDataStoreBase.get_relative_identifier)
    def get_relative_identifier(self, identifier):
        if isinstance(identifier, DataStoreMember) and identifier.parent is self:
            return identifier

        identifier = Path(identifier)
        identifier = identifier.name
        return identifier

    def open(self, identifier):
        if getattr(identifier, "parent", None) is not self:
            member = self.get_member(identifier)
        else:
            member = identifier

        _, record, _ = load_record_from_json(self.db.get(doc_id=member.id))
        return record

    def read(self, identifier):
        data = self.open(identifier)
        if self._md5 and isinstance(data, str):
            self._checksums[identifier] = get_text_hexdigest(data)

        return data

    @extend_docstring_from(ReadOnlyDataStoreBase.md5)
    def md5(self, member, force=True):
        md5_setting = self._md5  # for restoring automatic md5 calc setting
        if not getattr(member, "id", None):
            member = self.filtered(member)[0]

        if force and member not in self._checksums:
            self._md5 = True
            _ = member.read()

        result = self._checksums.get(member, None)
        self._md5 = md5_setting
        return result

    @property
    def logs(self):
        """returns all records with a .log suffix"""
        logfiles = []
        query = Query().identifier.matches(translate("*.log"))
        for record in self.db.search(query):
            member = DataStoreMember(record["identifier"], self, id=record.doc_id)
            logfiles.append(member)
        return logfiles

    @property
    def summary_logs(self):
        """returns a table summarising log files"""
        rows = []
        for record in self.logs:
            data = record.read().splitlines()
            first = data.pop(0).split("\t")
            row = [first[0], record.name]
            key = None
            mapped = {}
            for line in data:
                line = line.split("\t")[-1].split(" : ", maxsplit=1)
                if len(line) == 1:
                    mapped[key] += line[0]
                    continue

                key = line[0]
                mapped[key] = line[1]

            data = mapped
            row.extend(
                [
                    data["python"],
                    data["user"],
                    data["command_string"],
                    data["composable function"],
                ]
            )
            rows.append(row)
        return Table(
            header=["time", "name", "python version", "who", "command", "composable"],
            data=rows,
            title="summary of log files",
        )

    @property
    def describe(self):
        """returns tables describing content types"""
        lock_id = _db_lockid(self.source)
        if lock_id:
            title = (
                f"Locked db store. Locked to pid={lock_id}, current pid={os.getpid()}"
            )
        else:
            title = "Unlocked db store."
        num_incomplete = len(self.incomplete)
        num_complete = len(self.members)
        num_logs = len(self.logs)
        return Table(
            header=["record type", "number"],
            data=[
                ["completed", num_complete],
                ["incomplete", num_incomplete],
                ["logs", num_logs],
            ],
            title=title,
        )


class WritableTinyDbDataStore(
    ReadOnlyTinyDbDataStore, WritableDataStoreBase
):  # pragma: no cover
    @extend_docstring_from(WritableDirectoryDataStore.__init__)
    def __init__(self, *args, **kwargs):
        """

        Notes
        -----
        A TinyDb file can be locked. In which case, ``if_exists=OVERWRITE``
        will be converted to RAISE.
        """
        from cogent3.util.warning import discontinued

        discontinued(
            "class",
            "WritableTinyDbDataStore",
            "2022.3",
            "use sqlitedb via cogent3.open_data_store",
        )

        if_exists = kwargs.pop("if_exists", RAISE)
        create = kwargs.pop("create", True)
        ReadOnlyTinyDbDataStore.__init__(self, *args, **kwargs)
        WritableDataStoreBase.__init__(self, if_exists=if_exists, create=create)

    def _source_create_delete(self, if_exists, create):
        if not is_master_process():
            return

        path = Path(self.source)
        if if_exists == OVERWRITE and path.exists():
            try:
                path.unlink()
            except PermissionError:
                # probably user accidentally created a directory
                shutil.rmtree(path)
            return

        locked_id = _db_lockid(self.source)
        pid = os.getpid()
        if path.exists() and if_exists == RAISE:
            msg = f"'{path}' exists"
            if locked_id is not None:
                msg = (
                    f"{msg}, and is locked by process pid {locked_id}."
                    f" Current pid is {pid}"
                )
            raise FileExistsError(msg)

        if if_exists == IGNORE and locked_id is not None:
            warn(f"'{self.source}' is locked to {locked_id}, current pid is {pid}.")
            return

        if path.parent and not path.parent.exists() and not create:
            raise FileNotFoundError(f"'{path.parent}' does not exist, set create=True")

        path.parent.mkdir(parents=True, exist_ok=True)

    @extend_docstring_from(WritableDataStoreBase.write)
    def write(self, identifier, data):
        # writing into a tinydb has its own logic for conversion to json
        #  so we don't validate data is a string for this case
        from cogent3.app.composable import NotCompleted

        if isinstance(data, NotCompleted):
            return self.write_incomplete(identifier, data)

        super().write(identifier, "")
        id_suffix = identifier.split(".")[-1]
        if id_suffix not in (self.suffix, "log"):
            raise ValueError(
                f"identifier does not end with required suffix {self.suffix}"
            )

        matches = self.filtered(identifier)
        if matches:
            return matches[0]

        relative_id = self.get_relative_identifier(identifier)
        record = make_record_for_json(relative_id, data, True)
        doc_id = self.db.insert(record)

        member = DataStoreMember(relative_id, self, id=doc_id)
        if relative_id.endswith(self.suffix):
            self._members.append(member)

        return member

    def write_incomplete(self, identifier, not_completed):
        """stores an incomplete result object"""

        matches = self.filtered(identifier)
        if matches:
            return matches[0]

        relative_id = self.get_relative_identifier(identifier)
        record = make_record_for_json(relative_id, not_completed, False)
        doc_id = self.db.insert(record)

        return DataStoreMember(relative_id, self, id=doc_id)

    def add_file(self, path, make_unique=True, keep_suffix=True, cleanup=False):
        """
        Parameters
        ----------
        path : str
            location of file to be added to the data store
        keep_suffix : bool
            new path will retain the suffix of the provided file
        make_unique : bool
            a successive number will be added to the name before the suffix
            until the name is unique
        cleanup : bool
            delete the original
        """
        relativeid = self.make_relative_identifier(path)
        relativeid = Path(relativeid)
        path = Path(path)
        if keep_suffix:
            relativeid = str(relativeid).replace(
                relativeid.suffix, "".join(path.suffixes)
            )
            relativeid = Path(relativeid)

        if relativeid.suffix:
            name_wo_suffix = ".".join(relativeid.name.split(".")[:-1])
        else:
            name_wo_suffix = relativeid
        suffixes = "".join(relativeid.suffixes)
        query = Query().identifier.matches(f"{name_wo_suffix}*")
        num = self.db.count(query)
        if num:
            num += 1
            relativeid = str(relativeid).replace(suffixes, f"-{num}{suffixes}")
            relativeid = str(relativeid).replace(suffixes, f"-{num}{suffixes}")

        data = path.read_text()
        m = self.write(str(relativeid), data)

        if cleanup:
            path.unlink()

        return m



#!/usr/bin/env python

from unittest import TestCase, main

from cogent3.parse.greengenes import (
    MinimalGreengenesParser,
    SpecificGreengenesParser,
    make_ignore_f,
)


__author__ = "Daniel McDonald"
__copyright__ = "Copyright 2007-2022, The Cogent Project"  # consider project name
# remember to add yourself if you make changes
__credits__ = ["Daniel McDonald"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Daniel McDonald"
__email__ = "daniel.mcdonald@colorado.edu"
__status__ = "Prototype"


class ParseGreengenesRecordsTests(TestCase):
    def setUp(self):
        pass

    def test_MinimalGreengenesParser_mock(self):
        """Test MinimalGreengenesParser against mock data"""
        res = MinimalGreengenesParser(
            mock_data.splitlines(), RecStart="my_starting", RecEnd="my_ending"
        )

        records = list(res)

        exp = [
            {"a": "1", "b": "2", "c": "3", "d": "", "e": "5"},
            {"q": "asdasd", "c": "taco"},
        ]

        self.assertEqual(records, exp)

    def test_MinimalGreengenesParser_real(self):
        """Test MinimalGreengenesParser against real data"""
        res = MinimalGreengenesParser(real_data.splitlines())
        record1, record2 = list(res)

        self.assertEqual(record1["G2_chip_tax_string"], "Unclassified")
        self.assertEqual(
            record1["authors"],
            "Hernanandez-Eugenio,G., Silva-Rojas,H.V., Zelaya-Molina,L.X.",
        )
        self.assertEqual(record1["bel3_div_ratio"], "")
        self.assertEqual(len(record1), 72)

        self.assertEqual(record2["ncbi_acc_w_ver"], "FJ832719.1")
        self.assertEqual(record2["timestamp"], "2010-03-23 14:08:27")
        self.assertEqual(
            record2["title"],
            "Developmental Microbial Ecology of the Crop of the Folivorous Hoatzin",
        )

    def test_SpecificGreengenesParser_real(self):
        """Test SpecificGreengenesParser against real data"""
        fields = ["prokMSA_id", "journal"]
        res = SpecificGreengenesParser(real_data.splitlines(), fields)
        records = list(res)
        exp = [("604868", ""), ("604867", "ISME J (2010) In press")]
        self.assertEqual(records, exp)

        ids = ["604867", "12312312323"]
        res = SpecificGreengenesParser(real_data.splitlines(), fields, ids)
        records = list(res)
        exp = [("604867", "ISME J (2010) In press")]
        self.assertEqual(records, exp)

    def test_make_ignore_f(self):
        """Properly ignore empty records and the start line"""
        f = make_ignore_f("testing")
        self.assertFalse(f(["asasdasd", ""]))
        self.assertFalse(f(["test", ""]))
        self.assertFalse(f(["testing2", ""]))
        self.assertFalse(f(["testing", "asd"]))
        self.assertTrue(f(["", ""]))
        self.assertTrue(f(None))
        self.assertTrue(f(["", ""]))
        self.assertTrue(f(["testing", ""]))


mock_data = """my_starting
a=1
b=2
c=3
d=
e=5
my_ending

my_starting
q=asdasd
c=taco
my_ending
"""

real_data = """BEGIN
G2_chip_tax_string=Unclassified
G2_chip_tax_string_format_2=Unclassified
HOMD_tax_string=
HOMD_tax_string_format_2=
Hugenholtz_tax_string=Unclassified
Hugenholtz_tax_string_format_2=Unclassified
Ludwig_tax_string=Unclassified
Ludwig_tax_string_format_2=Unclassified
Pace_tax_string=Unclassified
Pace_tax_string_format_2=Unclassified
RDP_tax_string=Unclassified
RDP_tax_string_format_2=Unclassified
Silva_tax_string=Unclassified
Silva_tax_string_format_2=Unclassified
authors=Hernanandez-Eugenio,G., Silva-Rojas,H.V., Zelaya-Molina,L.X.
bel3_div_ratio=
bellerophon=
blast_perc_ident_to_template=
clone=51a 
contact_info=Irrigacion, Universidad Autonoma Chapingo, Carretera Mexico-Texcoco Km 37.5, Texcoco, Mexico 56230, Mexico
core_set_member=
core_set_member2=
country=Mexico: Mexico City 
create_date=21-NOV-2009
db_name=
decision=clone
description=Uncultured bacterium clone 51a 16S ribosomal RNA gene, partial sequence
email=
gold_id=
img_oid=
isolate=
isolation_source=mesophilic anaerobic reactor fed with effluent from the chemical industry 
journal=
longest_insertion=
medline_ids=
ncbi_acc=
ncbi_acc_w_ver=FJ461956.1
ncbi_gi=213390944
ncbi_seq_length=1512
ncbi_tax_id=77133
ncbi_tax_string=Bacteria; environmental samples
ncbi_tax_string_format_2=Unclassified
non_ACGT_count=
non_ACGT_percent=
note=
organism=uncultured bacterium
perc_ident_to_invariant_core=
prokMSA_id=604868
prokMSAname=Microbial ecology industrial digestor mesophilic anaerobic reactor fed effluent chemical industry clone 51a
pubmed_ids=
remark=
replaced_by=
single_nt_runs_over_7=
small_gap_intrusions=
source=uncultured bacterium
span_aligned=1..2
specific_host=
status=0
strain=
study_id=38002
sub_species=
submit_date=24-OCT-2008
template=
timestamp=2010-03-23 14:08:27
title=Microbial ecology of industrial anaerobic digestor
unaligned_length=
update_date=21-NOV-2009
warning=
wigeon95=
wigeon99=
wigeon_std_dev=
aligned_seq=unaligned
END

BEGIN
G2_chip_tax_string=Unclassified
G2_chip_tax_string_format_2=Unclassified
HOMD_tax_string=
HOMD_tax_string_format_2=
Hugenholtz_tax_string=Unclassified
Hugenholtz_tax_string_format_2=Unclassified
Ludwig_tax_string=Unclassified
Ludwig_tax_string_format_2=Unclassified
Pace_tax_string=Unclassified
Pace_tax_string_format_2=Unclassified
RDP_tax_string=Unclassified
RDP_tax_string_format_2=Unclassified
Silva_tax_string=Unclassified
Silva_tax_string_format_2=Unclassified
authors=Brodie,E.L., Dominguez-Bello,M.G., Garcia-Amado,M.A., Godoy-Vitorino,F., Goldfarb,K.C., Michelangeli,F.
bel3_div_ratio=
bellerophon=
blast_perc_ident_to_template=
clone=J3Q101_11C02 
contact_info=Biology, University of Puerto Rico, Rio Piedras Campus, PO Box 23360, San Juan, PR 00931-3360, USA
core_set_member=
core_set_member2=
country=Venezuela 
create_date=10-DEC-2009
db_name=
decision=clone
description=Uncultured bacterium clone J3Q101_11C02 16S ribosomal RNA gene, partial sequence
email=
gold_id=
img_oid=
isolate=
isolation_source=crop contents 
journal=ISME J (2010) In press
longest_insertion=
medline_ids=
ncbi_acc=
ncbi_acc_w_ver=FJ832719.1
ncbi_gi=226447371
ncbi_seq_length=1326
ncbi_tax_id=77133
ncbi_tax_string=Bacteria; environmental samples
ncbi_tax_string_format_2=Unclassified
non_ACGT_count=
non_ACGT_percent=
note=
organism=uncultured bacterium
perc_ident_to_invariant_core=
prokMSA_id=604867
prokMSAname=Microbial Ecology Crop Folivorous Hoatzin crop contents clone J3Q101_11C02
pubmed_ids=
remark=
replaced_by=
single_nt_runs_over_7=
small_gap_intrusions=
source=uncultured bacterium
span_aligned=1..2
specific_host=
status=0
strain=
study_id=37901
sub_species=
submit_date=16-MAR-2009
template=
timestamp=2010-03-23 14:08:27
title=Developmental Microbial Ecology of the Crop of the Folivorous Hoatzin
unaligned_length=
update_date=10-DEC-2009
warning=
wigeon95=
wigeon99=
wigeon_std_dev=
aligned_seq=unaligned
END
"""

if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Writer for GDE sequence format
"""

from cogent3.format.util import _AlignmentFormatter


__author__ = "Thomas La"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Jeremy Widmann", "Rob Knight", "Gavin Huttley", "Thomas La"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Thomas La"


def alignment_to_gde(alignment_dict, block_size=60, order=None):
    """Returns a Gde string given an alignment."""
    return GDEFormatter().format(
        alignment_dict, block_size, [] if order is None else order
    )


class GDEFormatter(_AlignmentFormatter):
    def format(self, alignment_dict, block_size, order):
        """Format the alignment to GDE.

        Parameters
        ----------
        alignment_dict
            dict of seq_name
        block_size
            the sequence length to write to each line,
            default is 60
        order
            optional list of sequence names, which order to
            print in.

        (Assumes complete and correct list of names)
        """

        # setup
        if not order:
            order = list(alignment_dict.keys())
        self.set_align_info(alignment_dict, order)
        self.set_block_size(block_size)

        return "".join(
            [
                "%s%s\n%s"
                % (
                    "%",
                    seq,
                    self.wrap_string_to_block_size(
                        alignment_dict[seq], alt_block_size=block_size
                    ),
                )
                for seq in self.align_order
            ]
        )



from __future__ import annotations

from collections import deque
from dataclasses import InitVar, dataclass, field
from itertools import product
from typing import Generator, Optional

import cogent3.util.progress_display as UI

from cogent3.core.sequence import Sequence


__author__ = "Gavin Huttley"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Gavin Huttley"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Alpha"


# deprecated code block
@UI.display_wrap
def dotplot(
    seq1, seq2, window, threshold, min_gap_length=0, band=None, ui=None
):  # pragma: no cover
    """A list of line segments covering the window-mers with identical matches > threshold

    gaps of size less than min_gap will be hidden, which saves on line segments.
    if 'band' is not None then it limits the searched area
    """
    from cogent3.util.warning import discontinued

    from . import compare_numba

    discontinued(
        "function",
        "dotplot",
        "2023.5",
        "replaced by much faster find_matched_paths",
    )

    segments_from_diagonal = compare_numba.segments_from_diagonal

    def one_diagonal(dia):
        segs = segments_from_diagonal(
            seq1, seq2, window, threshold, min_gap_length, dia
        )
        return [((start, start + dia), (end, end + dia)) for (start, end) in segs]

    if band is None:
        band = max(len(seq1), len(seq2))

    if isinstance(seq1, str):
        seq1 = seq1.encode("utf8")

    if isinstance(seq2, str):
        seq2 = seq2.encode("utf8")

    diagonals = list(range(-min(len(seq1), band), min(len(seq2), band) + 1))
    result = []
    for diag_segments in ui.imap(one_diagonal, diagonals, noun="offset"):
        result.extend(diag_segments)
    return result


@dataclass
class segment:
    """coordinates of a segment of a linear series"""

    start: int
    end: int

    def __contains__(self, index: int):
        return self.start <= index < self.end

    def __len__(self):
        return abs(self.end - self.start)

    def __sub__(self, other):
        a, b = (self, other) if self.start < other.start else (other, self)
        if a.overlap(b) or a.end == b.start:
            return segment(0, 0)

        return segment(a.end, b.start)

    def __nonzero__(self):
        return self.start != self.end

    def __getitem__(self, index: int):
        return (self.start, self.end)[index]

    def overlap(self, other):
        """whether coordinates overlap"""
        return (self.start < other.end <= self.end) or (
            self.start <= other.start < self.end
        )

    def adjacent(self, o):
        """whether coordinates are adjacent"""
        if not all([self, o]):
            return False
        return (o.start == self.end < o.end) or (self.start == o.end < self.end)

    def __or__(self, other):
        return self.merge(other, strict=True)

    def merge(self, other, strict: bool = True):
        """merge segments

        Parameters
        ----------
        other : segment
            instance to be merged
        strict
            if True, start
        Returns
        -------

        """
        if strict and not self.overlap(other):
            raise AssertionError(
                f"failed strict merge as {self} and {other} do not overlap"
            )
        return segment(min(self.start, other.start), max(self.end, other.end))

    def for_rc(self, length: int):
        """returns segment for reverse complement

        Parameters
        ----------
        length : int
            length of the sequence
        """
        return segment(length - self.end, length - self.start)


def _extend_left(
    matches: deque,
    seq1: str,
    idx1: int,
    seq2: str,
    idx2: int,
    left_limit: int,
    threshold: int,
    total: int,
) -> tuple[int, int, deque]:
    """extend a matching seed to the left

    Parameters
    ----------
    matches : deque
        series of bool values satisfying threshold
    seq1 : str
        first sequence
    idx1 : int
        initial position within seq1
    seq2 : str
        second sequence
    idx2 : int
        initial position within seq2
    left_limit : int
        positive integer indicating maximum number of left-wards
        positions to evaluate
    threshold : int
        minimum number of matches
    total : int
        total number of True values in matches

    Returns
    -------
    tuple[int, int, deque]
        The left-most position start indices for seq1/seq2 that satisfy
        threshold. The deque is the values corresponding to that position.
    """
    if idx1 == 0 or idx2 == 0:
        return idx1, idx2, matches

    left_limit = min(idx1, idx2, left_limit)

    matches.reverse()  # because we seek start backwards from seed
    offset = 0
    for offset in range(-1, -left_limit - 1, -1):
        prev = matches[0]
        current = seq1[idx1 + offset] == seq2[idx2 + offset]
        total += current - prev
        if total < threshold:
            # revert to previous position
            idx1 += offset + 1
            idx2 += offset + 1
            break

        matches.append(current)
    else:
        idx1 += offset
        idx2 += offset

    matches.reverse()
    return idx1, idx2, matches


def _extend_from_position(
    seq1, idx1, seq2, idx2, window, threshold, canonical, left_limit=0
) -> tuple[segment, segment]:
    """extend a matching seed

    Parameters
    ----------
    seq1 : str
        first sequence
    idx1 : int
        initial position within seq1
    seq2 : str
        second sequence
    idx2 : int
        initial position within seq2
    left_limit : int
        positive integer indicating maximum number of left-wards
        positions to evaluate
    threshold : int
        minimum number of matches
    total : int
        total number of True values in matches

    Returns
    -------
    tuple[int, int, deque]
        The left-most position start indices for seq1/seq2 that satisfy
        threshold. The deque is the values corresponding to that position.
    """
    canonical = set(canonical)
    if len(seq1) - idx1 < window or len(seq2) - idx2 < window:
        return segment(0, 0), segment(0, 0)
    matched = [seq1[idx1 + i] == seq2[idx2 + i] for i in range(window)]
    total = sum(matched)
    if total < threshold:
        return segment(0, 0), segment(0, 0)

    matches = deque(
        matched,
        maxlen=window,
    )
    idx1, idx2, matches = _extend_left(
        matches, seq1, idx1, seq2, idx2, left_limit, threshold, total
    )

    total = sum(matches)
    offset = -1
    for offset, (s1, s2) in enumerate(
        zip(seq1[idx1 + window :], seq2[idx2 + window :])
    ):
        prev = matches[0]
        matches.append(s1 == s2 if {s1, s2} < canonical else 0)
        total += matches[window - 1] - prev
        if total < threshold:
            break
    else:
        offset += 1

    return segment(idx1, idx1 + window + offset), segment(idx2, idx2 + window + offset)


@dataclass
class Kmer:
    """individual k-mer"""

    kmer: str
    ref_name: str
    index: InitVar[int]
    indices: dict = field(init=False)

    def __post_init__(self, index):
        self.kmer = str(self.kmer)
        self.indices = {self.ref_name: [index]}

    def __hash__(self) -> int:
        return hash(self.kmer)

    def __eq__(self, __o: object) -> bool:
        return getattr(__o, "kmer", __o) == self.kmer

    def add_location(self, seq_name: str, index: int):
        """add a location of self for seq_name

        Parameters
        ----------
        seq_name : str
            name of sequence
        index : int
            position of this k-mer

        Raises
        -------
        NotImplementedError if number of sequence names exceeds 2
        """
        indices = self.indices.get(seq_name, [])
        indices.append(index)
        self.indices[seq_name] = indices
        if len(self.indices) > 2:
            raise NotImplementedError(
                f"Maximum of 2 sequences allowed, have {list(self.indices.keys())}"
            )


@dataclass
class SeqKmers:
    """representing a sequence as Kmer instances"""

    seq: InitVar[Sequence]
    k: int
    canonical: set

    kmers: dict = field(init=False)
    num_seqs: int = field(init=False)
    ref_name: str = field(init=False)
    other_name: Optional[str] = field(init=False)

    def __post_init__(self, seq):
        self.canonical = set(self.canonical)
        canonical = self.canonical
        self.ref_name = name = seq.name

        kmers = {}
        for i, kmer in enumerate(seq.iter_kmers(k=self.k)):
            if not set(kmer) <= canonical:
                # exclude k-mers with  non-canonical characters
                continue

            if kmer in kmers:
                kmers[kmer].add_location(name, i)
            else:
                kmer = Kmer(kmer, name, i)
                kmers[kmer] = kmer

        self.kmers = kmers
        self.num_seqs = 1

    def add_seq(self, seq: Sequence) -> None:
        """transforms seq into k-mers, adds indices of matches

        Parameters
        ----------
        seq : Sequence
            has a different name to current sequence

        Notes
        -----
        k-mers unique to this sequence are ignored.
        """
        if seq.name == self.ref_name:
            raise ValueError(f"seq name {seq.name} matches ref seq name")

        self.other_name = name = seq.name

        for i, kmer in enumerate(seq.iter_kmers(self.k)):
            if kmer in self.kmers:
                self.kmers[kmer].add_location(name, i)

        self.num_seqs += 1

    def drop_seq(self, seq_name: str) -> None:
        """removes other seq from all k-mers"""
        for kmer in self.kmers:
            kmer.indices.pop(self.other_name, None)

        self.other_name = None
        self.num_seqs = 1

    def _iter_indices_1_seq(self) -> Generator:
        for kmer in self.iter_matching_kmers():
            indices = kmer.indices[self.ref_name]
            yield from product(indices, indices)

    def _iter_indices_2_seq(self) -> Generator:
        for kmer in self.iter_matching_kmers():
            indices = list(kmer.indices.values())
            yield from product(*indices)

    def iter_matched_indices(self) -> Generator:
        """generate (seq1 index, seq2 index) for matched k-mer seeds

        Notes
        -----

        The indices correspond to the start of occurrences of exact k-mer
        matches. If only one sequence, the indices are for matches off
        the main diagonal.
        """
        return (
            self._iter_indices_1_seq()
            if self.num_seqs == 1
            else self._iter_indices_2_seq()
        )

    def _iter_kmers_1_seq(self) -> Generator:
        for kmer in self.kmers:
            if len(kmer.indices[self.ref_name]) == 1:
                continue
            yield kmer

    def _iter_kmers_2_seq(self) -> Generator:
        for kmer in self.kmers:
            if len(kmer.indices) == 1:
                continue
            yield kmer

    def iter_matching_kmers(self) -> Generator:
        """generator of Kmer instances

        Notes
        -----
        If two sequences recorded, yields k-mers with entries in both
        sequences.
        If a single sequence, yields k-mers with multiple occurrences in the
        primary sequence.
        """
        return (
            self._iter_kmers_1_seq() if self.num_seqs == 1 else self._iter_kmers_2_seq()
        )


@dataclass
class MatchedSeqPaths:
    """holder of paths between two sequences"""

    name: str = ""
    paths: dict = field(init=False)

    def __post_init__(self):
        self.paths = {}

    def __getitem__(self, y_intercept: int):
        self.paths[y_intercept] = self.paths.get(y_intercept, [])
        return self.paths[y_intercept]

    def last_on_path(self, refseq_pos: int, otherseq_pos: int):
        """returns the last coordinates on path"""
        y_intercept = otherseq_pos - refseq_pos
        if y_intercept in self.paths:
            return self[y_intercept][-1]

        # if no entry, corresponds to origin for refseq, y_intercept for other
        return segment(0, 0), segment(y_intercept, y_intercept)

    def append(self, seq1_segment: segment, seq2_segment: segment) -> None:
        """
        add segment to path

        Parameters
        ----------
        seq1_segment,  seq2_segment
            spans for refseq and otherseq

        Notes
        -----
        If a segment has 0 distance to the last segment, they are merged
        """
        y_intercept = seq2_segment.start - seq1_segment.start
        last_x, last_y = self.last_on_path(seq1_segment.start, seq2_segment.start)
        if seq1_segment.adjacent(last_x):
            self[y_intercept][-1] = (
                last_x.merge(seq1_segment, strict=False),
                last_y.merge(seq2_segment, strict=False),
            )
        else:
            self[y_intercept].append((seq1_segment, seq2_segment))

    def position_covered(self, refseq_pos: int, otherseq_pos: int) -> bool:
        """evaluates whether the position already evaluated for y_intercept"""
        evaluated = self.paths.get(otherseq_pos - refseq_pos, [])
        # I'm reversing the order here in the expectation that the most
        # recent evaluation is most likely to cover a span
        return any(
            refseq_pos in refseq_segment for refseq_segment, _ in evaluated[::-1]
        )

    def _get_segments(self, min_gap):
        if not min_gap:
            return self.paths

        paths = {}
        for y_intercept in self.paths:
            result = self.paths[y_intercept][:]
            if len(result) == 1:
                paths[y_intercept] = result
                continue
            result = result[:1]
            for x, y in self.paths[y_intercept][1:]:
                last_x, last_y = result[-1]
                if len(x - last_x) <= min_gap:
                    result[-1] = last_x.merge(x, strict=False), last_y.merge(
                        y, strict=False
                    )
                else:
                    result.append((x, y))
            paths[y_intercept] = result
        return paths

    def get_coords(
        self, rc: bool = False, length: Optional[int] = None, min_gap: int = 0
    ):
        """returns x, y coordinates for plotting

        Parameters
        ----------
        rc
            whether coordinates are for a reverse complemented sequence
        length
            length of the sequence
        min_gap
            connect disjoint segments separated by <= min_gap
        """
        paths = self._get_segments(min_gap)
        x = []
        y = []
        for y_intercept in paths.values():
            for a, b in y_intercept:
                if rc:
                    # adjust coordinate and reverse
                    s, e = b.for_rc(length)
                    b = segment(e, s)
                x.extend(a)
                x.append(None)
                y.extend(b)
                y.append(None)
        return x[:-1], y[:-1]  # discard last None

    def plotly_trace(
        self, rc: bool = False, length: Optional[int] = None, min_gap: int = 0
    ):
        x, y = self.get_coords(rc=rc, length=length, min_gap=min_gap)
        trace = {
            "type": "scatter",
            "x": x,
            "y": y,
            "mode": "lines",
        }
        if self.name:
            trace["name"] = self.name
            trace["showlegend"] = True
        return trace


def _calc_seed_size(w: int, t: int) -> int:
    """computes k-mer size

    Parameters
    ----------
    w : int
        window size
    t : int
        threshold for minimum number of matches

    Returns
    -------
    int
        returns the larger of (w // k) * k <= t OR w - t + 1
    """
    assert 0 < t <= w, f"threshold={t} > window size={w}"

    # k should be such that (w % k) * k < threshold
    denom = w - t + 1
    return max(min(w // denom, t), denom)


def find_matched_paths(
    seq_kmers: SeqKmers,
    seq1: Sequence,
    seq2: Optional[Sequence] = None,
    window: int = 20,
    threshold: int = 17,
) -> MatchedSeqPaths:
    """determine all matches between seq1 and seq21

    Parameters
    ----------
    seq_kmers : SeqKmers
        instance with seq1 converted to k-mers
    seq1, seq2 : Sequence
        cogent3 Sequence instances. If seq2 not provided, compares seq1 to
        itself.
    window : int
        size of sequence segment to be compared
    threshold : int
        Minimum number of positions that must be equal within the window.Less than or equal to window.

    Returns
    -------
    MatchedSeqPaths
    """
    k = seq_kmers.k
    delta = (window // k) * (k - 1) + (window - threshold)
    canonical = set(seq1.moltype)
    paths = MatchedSeqPaths()

    if seq2 is None:
        seq2 = seq1
        paths[0].append((segment(0, len(seq1)), segment(0, len(seq1))))
    else:
        seq_kmers.add_seq(seq2)

    # for every matched k-mer, we determine the y-intercept.
    for seq1_idx, seq2_idx in seq_kmers.iter_matched_indices():
        if paths.position_covered(seq1_idx, seq2_idx):
            # seq1_idx already evaluated for y_intercept, skip
            continue

        # it hasn't been evaluated, compute the left most limit to start from
        s1_coord, _ = paths.last_on_path(seq1_idx, seq2_idx)
        left_limit = min(abs(s1_coord.end - seq1_idx), delta)
        ref_coord, other_coord = _extend_from_position(
            seq1._seq,
            seq1_idx,
            seq2._seq,
            seq2_idx,
            window=window,
            threshold=threshold,
            canonical=canonical,
            left_limit=left_limit,
        )

        if not ref_coord:
            continue

        paths.append(ref_coord, other_coord)

    return paths



#!/usr/bin/env Python
"""Data for molecular weight calculations on proteins and nucleotides."""

__author__ = "Rob Knight"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Rob Knight"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Production"

ProteinWeights = {
    "A": 71.09,
    "C": 103.16,
    "D": 115.10,
    "E": 129.13,
    "F": 147.19,
    "G": 57.07,
    "H": 137.16,
    "I": 113.18,
    "K": 128.19,
    "L": 113.18,
    "M": 131.21,
    "N": 114.12,
    "P": 97.13,
    "Q": 128.15,
    "R": 156.20,
    "S": 87.09,
    "T": 101.12,
    "V": 99.15,
    "W": 186.23,
    "Y": 163.19,
    "U": 150.06,
}

RnaWeights = {"A": 313.21, "U": 290.17, "C": 289.19, "G": 329.21}

DnaWeights = {"A": 297.21, "T": 274.17, "C": 273.19, "G": 313.21}

ProteinWeightCorrection = 18.0  # terminal residues not dehydrated
DnaWeightCorrection = 61.96  # assumes 5' monophosphate, 3' OH
RnaWeightCorrection = DnaWeightCorrection


class WeightCalculator(object):
    """Calculates molecular weight of a non-degenerate sequence."""

    def __init__(self, Weights, Correction):
        """Returns a new WeightCalculator object (class, so serializable)."""
        self.Weights = Weights
        self.Correction = Correction

    def __call__(self, seq, correction=None):
        """Returns the molecular weight of a specified sequence."""
        if not seq:
            return 0
        if correction is None:
            correction = self.Correction
        get_mw = self.Weights.get
        return sum([get_mw(i, 0) for i in seq]) + correction


DnaMW = WeightCalculator(DnaWeights, DnaWeightCorrection)
RnaMW = WeightCalculator(RnaWeights, DnaWeightCorrection)
ProteinMW = WeightCalculator(ProteinWeights, ProteinWeightCorrection)



#!/usr/bin/env python
"""Tests for FASTA sequence format writer.
"""
from unittest import TestCase, main

from cogent3.core.alignment import Alignment
from cogent3.core.info import Info
from cogent3.core.sequence import Sequence
from cogent3.format.fasta import alignment_to_fasta


__author__ = "Jeremy Widmann"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Jeremy Widmann", "Gavin Huttley", "Rob Knight"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Jeremy Widmann"
__email__ = "jeremy.widmann@colorado.edu"
__status__ = "Production"


class FastaTests(TestCase):
    """Tests for Fasta writer."""

    def setUp(self):
        """Setup for Fasta tests."""
        self.strings = ["AAAA", "CCCC", "gggg", "uuuu"]
        self.labels = ["1st", "2nd", "3rd", "4th"]
        self.infos = ["Dog", "Cat", "Mouse", "Rat"]
        self.sequences_with_labels = list(map(Sequence, self.strings))
        self.sequences_with_names = list(map(Sequence, self.strings))
        for l, sl, sn in zip(
            self.labels, self.sequences_with_labels, self.sequences_with_names
        ):
            sl.label = l
            sn.name = l
        self.fasta_no_label = ">0\nAAAA\n>1\nCCCC\n>2\ngggg\n>3\nuuuu\n"
        self.fasta_with_label = ">1st\nAAAA\n>2nd\nCCCC\n>3rd\nGGGG\n>4th\nUUUU\n"
        self.fasta_with_label_lw2 = (
            ">1st\nAA\nAA\n>2nd\nCC\nCC\n>3rd\nGG\nGG\n>4th\nUU\nUU\n"
        )
        self.alignment_dict = {
            "1st": "AAAA",
            "2nd": "CCCC",
            "3rd": "GGGG",
            "4th": "UUUU",
        }
        self.alignment_object = Alignment(self.alignment_dict)
        for label, info in zip(self.labels, self.infos):
            self.alignment_object.named_seqs[label].info = Info(species=info)
        self.fasta_with_label_species = (
            ">1st:Dog\nAAAA\n>2nd:Cat\nCCCC\n>3rd:Mouse\nGGGG\n>4th:Rat\nUUUU\n"
        )
        self.alignment_object.RowOrder = ["1st", "2nd", "3rd", "4th"]

    def test_alignment_to_fasta(self):
        """should return correct fasta string."""
        self.assertEqual(alignment_to_fasta({}), "")
        self.assertEqual(alignment_to_fasta(self.alignment_dict), self.fasta_with_label)
        self.assertEqual(
            alignment_to_fasta(self.alignment_dict, block_size=2),
            self.fasta_with_label_lw2,
        )


if __name__ == "__main__":
    main()
